"""Util functions.
"""
import os


def is_jsonl(name: str) -> bool:
    """Test if a filename is json lines.
    Also returns True for '<stdin>' and '<stdout>'.

    Arguments:
        name {str} -- name to test

    Returns:
        {bool} -- True if this name is jsonl.
    """
    if name in ["<stdin>", "<stdout>"]:
        return True
    _, suffix = os.path.splitext(name)
    # print('suffix = {suffix}'.format(suffix=suffix))
    return suffix in [".jsonl"]



import io
import json
from string import printable

import pytest
import hypothesis
from hypothesis import given, settings, strategies as st

from sb_json_tools import json_iter, jsonl_iter
from sb_json_tools.tests.utils import (
    compare_iters,
    JSON_DATA,
    convert_to_decimal_if_float,
    reference_dict,
)


DATA = [
    {"a": 1},
    {"a": 2},
]

JSON_FACIT = '[\n{"a": 1},\n{"a": 2}\n]'
JSONL_FACIT = '{"a": 1}\n{"a": 2}\n'


def gen_data():
    for i in DATA:
        yield i


@pytest.mark.parametrize("out", [io.StringIO, io.BytesIO])
@pytest.mark.parametrize(
    "it,data", [(json_iter, DATA[0]), (jsonl_iter, DATA[0]),],
)
def test_dump_dict_memoryio(out, it, data):
    out = out()
    it.dump(data, out)

    out.seek(0)
    for i, obj in enumerate(it.load(out)):
        if i == 0:
            assert obj == data
        else:
            pytest.fail()


@pytest.mark.parametrize("out", [io.StringIO, io.BytesIO])
@pytest.mark.parametrize("it", [json_iter, jsonl_iter])
@given(st.just("00"))
def test_dump_strings_with_zeros_memoryio(out, it, data):
    out = out()
    it.dump(data, out)
    facit = '"00"'
    if isinstance(out, io.BytesIO):
        facit = facit.encode("utf-8")
    assert out.getvalue() == facit

    out.seek(0)
    for i, obj in enumerate(it.load(out)):
        if i == 0:
            assert obj == data
        else:
            pytest.fail()

    # out.seek(0)
    # if not isinstance(data, list):
    #     data = [data]
    # compare_iters(it.load(out), data)


@pytest.mark.parametrize("out", [io.StringIO, io.BytesIO])
@pytest.mark.parametrize("it", [json_iter, jsonl_iter])
@given(JSON_DATA)
@hypothesis.settings(suppress_health_check=[hypothesis.HealthCheck.too_slow])
def test_dump_array_memoryio(out, it, data):
    out = out()
    it.dump(data, out)

    facit = reference_dict(data)

    out.seek(0)
    if isinstance(data, list):
        compare_iters(it.load(out), facit)
    else:
        for i, obj in enumerate(it.load(out)):
            if i == 0:
                assert obj == facit
            else:
                pytest.fail()


@pytest.mark.parametrize("out", [io.StringIO, io.BytesIO])
@pytest.mark.parametrize(
    "it, data_generator", [(json_iter, gen_data), (jsonl_iter, gen_data),]
)
def test_dump_gen_memoryio(out, it, data_generator):
    out = out()
    it.dump(data_generator(), out)
    out.seek(0)
    compare_iters(it.load(out), data_generator())


@pytest.mark.parametrize("file_mode", ["rb", "r"])
@pytest.mark.parametrize(
    "it,file_name,facit", [(json_iter, "sb_json_tools/tests/data/array.json", None),]
)
def test_load_file_name(it, file_name: str, facit, file_mode):
    if not facit:
        facit = file_name
    with open(facit) as fp:
        facit_it = json.load(fp)
    test_it = it.load_from_file(file_name, file_mode=file_mode)
    compare_iters(test_it, facit_it)




"""Module to validate json object according to a json-schema.org schema."""

import fastjsonschema

from typing import Dict
from typing import Generator
from typing import Iterable
from typing import List
from typing import Tuple
from typing import Union

from . import exceptions


__version__ = '0.1.0'


class Result(object):
    def __init__(self,
                 ok=None,
                 error=None,
                 error_msg=None):
        if ok is not None and error is not None:
            raise ValueError()
        self.ok = ok
        self.error = error
        self.error_msg = error_msg


def ok(item: Dict) -> Tuple[Dict, None]:
    return (item, None)


def error(item: Dict, msg: str) -> Tuple[None, Dict]:
    return (None, {'failing': item, 'message': msg})


def streaming_validate(
        schema: Dict,
        data: Union[Dict, Iterable[Dict]],
        *,
        raise_on_error: bool = False
        ) -> Generator[Tuple, None, Tuple]:
    try:
        validator = fastjsonschema.compile(schema)
    except fastjsonschema.JsonSchemaDefinitionException as e:
        raise exceptions.SchemaDefinitionException(e.message)

    if isinstance(data, dict):
        try:
            valid_obj = validator(data)
            return ok(valid_obj)
        except fastjsonschema.JsonSchemaException as e:
            if raise_on_error:
                raise exceptions.ValidationException(e.message, data)
            return error(data, e.message)

    for orig_obj in data:
        try:
            valid_obj = validator(orig_obj)
        except fastjsonschema.JsonSchemaException as e:
            if raise_on_error:
                raise exceptions.ValidationException(e.message, orig_obj)
            yield error(orig_obj, e.message)
        else:
            yield ok(valid_obj)


def validate(
        schema: Dict,
        data: Union[Dict, Iterable[Dict]],
        *,
        raise_on_error: bool = False
        ) -> Tuple[List[Dict], List[Dict]]:
    errors = []
    correct = []

    for ok, error in streaming_validate(schema, data, raise_on_error=raise_on_error):
        if ok:
            correct.append(ok)
        else:
            errors.append(error)
    return correct, errors


def processing_validate(
        schema: Dict,
        data: Union[Dict, Iterable[Dict]],
        *,
        on_ok: Generator[None, Dict, None],
        on_error: Generator[None, Dict, None]
        ) -> None:
    for ok, error in streaming_validate(schema, data):
        if ok:
            on_ok.send(ok)
        else:
            on_error.send(error)



import io

import pytest

import hypothesis
from hypothesis import given, strategies as st

from sb_json_tools import jt_iter, jsonlib

from sb_json_tools.tests.utils import (
    compare_iters,
    JSON_DATA,
    convert_to_decimal_if_float,
    reference_dict,
)


@pytest.mark.parametrize("out", [io.StringIO, io.BytesIO])
@pytest.mark.parametrize(
    "file_type", ["json", "jsonl"],
)
@given(JSON_DATA)
@hypothesis.settings(suppress_health_check=[hypothesis.HealthCheck.too_slow])
def test_dump_and_load_memoryio(out, file_type, data):
    out = out()
    jt_iter.dump(data, out, file_type=file_type)
    facit = reference_dict(data)

    out.seek(0)  # read it from the beginning
    if isinstance(data, list):
        out_iter = jt_iter.load(out, file_type=file_type)
        compare_iters(out_iter, facit)
    else:
        for i in jt_iter.load(out, file_type=file_type):
            assert i == facit



from sb_json_tools import jt_val as jv


schema = {
    "type": "object",
    "properties": {
        "valid": {
            "type": "integer"
        }
    },
    "required": ["valid"]
}


def gen_valid_data(n):
    for i in range(n):
        yield {'valid': i}


def gen_invalid_data(n):
    for i in range(n):
        yield {'invalid': i}


def gen_mixed_data(n):
    for i in range(n):
        if i % 2 == 1:
            yield {'valid': i}
        else:
            yield {'invalid': i}


def test_valid_data():
    n = 10
    stream = jv.streaming_validate(schema, gen_valid_data(n))

    for i, (ok, error) in enumerate(stream):
        assert ok
        assert ok['valid'] == i
        assert not error


def test_invalid_data():
    n = 10
    stream = jv.streaming_validate(schema, gen_invalid_data(n))
    for ok, error in stream:
        assert error
        assert not ok


def test_mixed_data():
    n = 10
    stream = jv.streaming_validate(schema, gen_mixed_data(n))
    for i, (ok, error) in enumerate(stream):
        if i % 2 == 1:
            assert ok
            assert not error
        else:
            assert error
            assert not ok



"""Common test functions."""
import decimal
import itertools
import math

# import decimal
from string import printable

from hypothesis import strategies as st

from sb_json_tools import jsonlib

JSON_DATA = st.recursive(
    st.none()
    | st.booleans()
    | st.floats(allow_nan=False, allow_infinity=False)
    | st.integers()
    | st.text(printable),
    lambda children: st.lists(children, 1).filter(lambda l: len(l) > 1 or l[0] is None)
    | st.dictionaries(
        st.text(printable).filter(lambda s: len(s) > 0), children, min_size=1
    ),
)


def reference_dict(x):
    """Create a reference dict for testing.

    Arguments:
        x {any} -- the object to make a reference object

    Returns:
        any -- the reference version
    """
    return jsonlib.loads(
        jsonlib.dumps(x, **jsonlib.JSON_SETTINGS), **jsonlib.JSON_SETTINGS
    )


def convert_to_decimal_if_float(x):
    """Convert value to float to Decimal.

    Arguments:
        x {any} -- The value to test and convet if float.

    Returns:
        any -- Decimal if x is a float, otherwise x
    """
    if isinstance(x, float):
        return decimal.Decimal(x)
    else:
        return x


def assert_equal_not_none(i1, i2):
    """Assert to objects are equal and no one is None.

    Arguments:
        i1 {any} -- first object
        i2 {any} -- second object
    """
    assert i1 is not None
    assert i2 is not None
    if isinstance(i1, (decimal.Decimal, float)):
        assert math.isclose(i1, i2)
    else:
        assert i1 == i2


def compare_iters(it1, it2):
    """Compare two iterables with asserts.

    Arguments:
        it1 {Iterable} -- first iterable
        it2 {Iterable} -- second iterable
    """
    for i1, i2 in itertools.zip_longest(it1, it2):
        if i1 is None and i1 == i2:
            continue
        else:
            if isinstance(i1, list) and isinstance(i2, list):
                compare_iters(i1, i2)
            else:
                assert_equal_not_none(i1, i2)



import os

import pytest

from click.testing import CliRunner

from sb_json_tools import cli


def _to_path(file_name):
    return os.path.join(os.path.dirname(__file__), file_name)


def _invoke_commad(schema, arg):
    return '--schema {0} {1}'.format(_to_path(schema), _to_path(arg))


@pytest.fixture
def runner():
    return CliRunner()


def test_cli(runner):
    result = runner.invoke(cli.cli)
    assert result.exit_code == 0
    assert not result.exception
    # assert result.output.strip() == 'Hello, world.'


def test_cli_with_short_option_no_parameter(runner):
    result = runner.invoke(cli.cli, ['-s'])
    assert result.exception
    assert result.exit_code == 2
    # assert result.output.strip() == 'Howdy, world.'


def test_cli_with_long_option_no_parameter(runner):
    result = runner.invoke(cli.cli, ['--schema'])
    assert result.exception
    assert result.exit_code == 2
    # assert result.output.strip() == 'Howdy, world.'


def test_cli_with_short_option_parameter(runner):
    result = runner.invoke(cli.cli, ['-s schema.json'])
    assert result.exception
    assert result.exit_code == 2
    # assert result.output.strip() == 'Howdy, world.'


def test_cli_with_long_option_parameter(runner):
    result = runner.invoke(cli.cli, ['--schema schema.json'])
    assert result.exception
    assert result.exit_code == 2
    # assert result.output.strip() == 'Howdy, world.'


@pytest.mark.xfail(reason='working on it')
def test_cli_with_schema_and_valid_arg(runner):
    schema = "sb_json_tools/tests/data/schema_default.json"
    arg = "sb_json_tools/tests/valid_array.json"
    result = runner.invoke(cli.cli, [
        'validate',
        '--schema',
        schema,
        arg,
        '-'
    ])
    print('result = {result}'.format(result=repr(result)))
    assert result.exit_code == 0
    assert not result.exception
    assert (
        result.output.strip()
        ==
        'Validating tests/{0} with the schema in tests/{1}.'.format(arg, schema)
    )


@pytest.mark.xfail(reason='working on it')
def test_cli_with_schema_and_invalid_type_arg(runner):
    schema = "tests/data/schema_default.json"
    arg = "tests/data/invalid_type.json"
    result = runner.invoke(cli.cli, ['--schema', schema, arg, '-'])
    assert result.exit_code == 0
    assert not result.exception
    assert result.output.strip() == 'Validating {0} with the schema in {1}.'.format(arg, schema)


@pytest.mark.xfail(reason='working on it')
def test_cli_with_schema_and_invalid_property_arg(runner):
    schema = "schema.json"
    arg = "invalid_property.json"
    result = runner.invoke(cli.cli, ['--schema', schema, arg, '-'])
    assert result.exit_code == 0
    assert not result.exception
    assert result.output.strip() == 'Validating {0} with the schema in {1}.'.format(arg, schema)



""" Diff two JSON-files (or dict's). """
from sb_json_tools import jsonlib

# Borrow from http://djangosnippets.org/snippets/2247/
# and inspired by https://github.com/jclulow/jsondiff
# Modified to return information more useful to us


TYPE = 'TYPE'
PATH = 'PATH'
VALUE = 'VALUE'


class Diff(object):
    def __init__(self, first, second, with_values=False):
        self.difference = []
        self.seen = []
        self.check(first, second, with_values=with_values)

    def check(self, first, second, path='', with_values=False):
        if with_values and second is not None:
            if not isinstance(first, type(second)):
                # Type and hence value change:
                self.save_diff(path, type(first).__name__,
                               type(second).__name__, TYPE)
                self.save_diff(path, first, second, PATH)

        if isinstance(first, dict):
            for key in first:
                # the first part of path must not have trailing dot.
                if len(path) == 0:
                    new_path = key
                else:
                    new_path = "%s.%s" % (path, key)

                if isinstance(second, dict):
                    if key in second:
                        sec = second[key]
                    else:
                        #  Add or remove:
                        #  there are key in the first,
                        # that is not presented in the second
                        self.save_diff(new_path, first.get(key), '', PATH)

                        # prevent further values checking.
                        sec = None

                    # recursive call
                    if sec is not None:
                        self.check(first[key], sec, path=new_path,
                                   with_values=with_values)

        # if object is list, loop over it and check.
        elif isinstance(first, list):
            for (index, item) in enumerate(first):
                new_path = "%s[%s]" % (path, index)
                # try to get the same index from second
                sec = None
                if second is not None:
                    try:
                        sec = second[index]
                    except (IndexError, KeyError):
                        # List with new member
                        self.save_diff(new_path, item, '', TYPE)

                # recursive call
                self.check(first[index], sec, path=new_path,
                           with_values=with_values)

        # not list, not dict. Check for equality (only if with_values is True)
        # and return.
        else:
            if with_values and second is not None:
                if first != second:
                    # Value change
                    self.save_diff(path, first, second, VALUE)
            return

    def save_diff(self, path, diff_before, diff_after, type_):
        long_message = '%s - %s | %s' % (path, diff_before, diff_after)
        if long_message not in self.seen:
            self.seen.append(long_message)
            self.difference.append((type_, path, diff_before, diff_after))


def get_content(location):
    if isinstance(location, dict):  # compatability with python3
        return location

    content = open(location, 'r').read()
    if content is None:
        raise Exception("Could not load content for " + location)
    return jsonlib.loads(content)


def compare_gen(location1, location2, print_all=False):
    """ Compare two JSON files (or dicts) and yield the result. """
    json1 = get_content(location1)
    json2 = get_content(location2)
    diff1 = Diff(json1, json2, True).difference
    diff2 = Diff(json2, json1, False).difference
    for type_, path, before, after in diff1:
        if before and after:
            action = "TYPECHANGE" if type_ == "TYPE" else "CHANGE"
            yield {'type': action, 'field': path, 'before': before, 'after': after}
        elif before:
            yield {'type': 'REMOVED', 'field': path, 'before': before}
        elif print_all:
            # changes should be caught in the first if clause, additions below
            yield {'type': 'CHANGE', 'field': path, 'after': after, 'before': before}
    for type_, path, after, before in diff2:
        yield {'type': 'ADDED', 'field': path, 'after': after}


def compare(location1, location2, print_all=False):
    """
    Compare two JSON files (or dicts).

    Wraps the generator version `compare_gen`.
    """
    return list(compare_gen(location1, location2, print_all=print_all))



"""
Utility library to load the underlying json library.

Imports `ujson` if it is present, otherwise it imports `json` from
the standard library.
"""
from simplejson import dump, dumps, load, loads

# try:
#     from ujson import dump, dumps, load, loads  # pylint: disable=unused-import
# except ImportError:
#     from json import dump, dumps, load, loads  # noqa: F401


JSON_SETTINGS = {"use_decimal": True}


def load_from_file(file_name: str):
    """
    Load the JSON file with the given file_name.

    :param file_name: name of the file to load from.
    :return: the loaded JSON file.
    """
    with open(file_name, "r") as fp:
        return load(fp)


def dump_to_file(obj, file_name: str):
    """
    Dump to a JSON file with the given file name.

    :param obj: the object to dump.
    :param file_name: name of the file to dump to.
    :return: anything returned from the backend.
    """
    with open(file_name, "w") as fp:
        return dump(obj, fp)



from string import printable

import pytest

from hypothesis import given, strategies as st

from sb_json_tools import jsondiff as diff




objs = [
    {'a': 1},  # 0
    {'a': 2},  # 1
    {'a': 1, 'b': 1},  # 2
    {'a': '1'},  # 3
    {'a': {'s': [1]}},  # 4
    {'a': {'s': [1, 2]}},  # 5
    {'a': {'s': '1'}},  # 6
]

indices = [i for i in range(len(objs))]


def change(field, before, after):
    return {'before': before, 'after': after, 'field': field, 'type': 'CHANGE'}


def typechange(field, before, after):
    return {
        'before': before,
        'after': after,
        'field': field,
        'type': 'TYPECHANGE'
        }


def added(field, after):
    return {'after': after, 'field': field, 'type': 'ADDED'}


def removed(field, before):
    return {'before': before, 'field': field, 'type': 'REMOVED'}


_data_test_compare = [
    (0, 1, [change('a', 1, 2)]),
    (0, 2, [added('b', 1)]),
    (0, 3, [typechange('a', 'int', 'str'), change('a', 1, '1')]),
    (0, 4, [typechange('a', 'int', 'dict'), change('a', 1, {'s': [1]})]),
    (1, 0, [change('a', 2, 1)]),
    (1, 2, [change('a', 2, 1), added('b', 1)]),
    (2, 0, [removed('b', 1)]),
    (2, 1, [change('a', 1, 2), removed('b', 1)]),
    (4, 5, [added('a.s[1]', 2)]),
    (4, 6, [
        typechange('a.s', 'list', 'str'),
        change('a.s', [1], '1'),
        typechange('a.s[0]', 'int', 'str'),
        change('a.s[0]', 1, '1'),
    ]),
]
for i in range(len(objs)):
    _data_test_compare.append((i, i, []))


@pytest.mark.parametrize('i1,i2,facit', _data_test_compare)
def test_compare(i1, i2, facit):
    r = diff.compare(objs[i1], objs[i2])

    assert len(r) == len(facit)

    for a in r:
        assert a in facit


@given(st.dictionaries(st.text(printable), st.none() | st.floats(allow_nan=False) | st.integers() | st.text(printable)))
def test_diff_equal(data):
    r = diff.compare(data, data)

    assert len(r) == 0




""" Handle JSON lazy. """
from typing import Dict
from typing import IO
from typing import Iterable
from typing import Union

import codecs
import io

import ijson

from sb_json_tools import jsonlib


def dump(data: Union[Dict, Iterable], fp: IO):
    """ Dump array to a file object.

    Parameters
    ----------
    fp :
        File object to write to. Must be writable.
    data :
        Iterable object to write.
    """
    if isinstance(fp, io.BufferedIOBase):
        fp = codecs.getwriter("utf-8")(fp)

    if isinstance(data, (dict, str, int, float, bool)):
        fp.write(jsonlib.dumps(data, **jsonlib.JSON_SETTINGS))
        return

    try:
        it = iter(data)
    except TypeError:
        fp.write(jsonlib.dumps(data, **jsonlib.JSON_SETTINGS))
        return

    fp.write("[\n")
    try:
        obj = next(it)
        fp.write(jsonlib.dumps(obj, **jsonlib.JSON_SETTINGS))
    except StopIteration:
        pass
    else:
        for v in it:
            fp.write(",\n")
            fp.write(jsonlib.dumps(v, **jsonlib.JSON_SETTINGS))
    fp.write("\n]")


def load(fp: IO) -> Iterable:
    yield from ijson.items(fp, "item")


def load_eager(fp: IO):
    data = jsonlib.load(fp, **jsonlib.JSON_SETTINGS)
    if isinstance(data, list):
        for obj in data:
            yield obj
    else:
        return data


def load_from_file(file_name: str, *, file_mode: str = None):
    if not file_mode:
        file_mode = "br"
    with open(file_name, file_mode) as fp:
        yield from load(fp)


def dump_to_file(gen: Iterable, file_name: str, *, file_mode: str = None):
    if not file_mode:
        file_mode = "w"
    with open(file_name, file_mode) as fp:
        return dump(gen, fp)



from sb_json_tools.utils import is_jsonl


def test_jsonl():
    file_name = "test.jsonl"
    assert is_jsonl(file_name)


def test_json():
    file_name = "test.json"
    assert not is_jsonl(file_name)



from typing import Iterable
from typing import IO

from . import json_iter
from . import jsonl_iter
from . import utils


def load(
    fp: IO,
    *,
    file_type=None
) -> Iterable:
    _iter = json_iter
    if file_type == 'json':
        pass
    elif file_type == 'jsonl':
        _iter = jsonl_iter
    else:
        if utils.is_jsonl(fp.name):
            _iter = jsonl_iter

    yield from _iter.load(fp)


def load_from_file(
        file_name: str,
        *,
        file_type: str = None,
        file_mode: str = None
        ):
    _iter = json_iter
    if file_type == 'json':
        pass
    elif file_type == 'jsonl':
        _iter = jsonl_iter
    else:
        if utils.is_jsonl(file_name):
            _iter = jsonl_iter

    yield from _iter.load_from_file(file_name, file_mode=file_mode)


def dump(
        in_iter_,
        fp: IO,
        *,
        file_type: str = None
        ):
    _iter = json_iter
    if file_type == 'json':
        pass
    elif file_type == 'jsonl':
        _iter = jsonl_iter
    else:
        if utils.is_jsonl(fp.name):
            _iter = jsonl_iter

    _iter.dump(in_iter_, fp)


def dump_to_file(
        in_iter_,
        file_name: str,
        *,
        file_type=None,
        file_mode: str = None
):
    _iter = json_iter
    if file_type == 'json':
        pass
    elif file_type == 'jsonl':
        _iter = jsonl_iter
    else:
        if utils.is_jsonl(file_name):
            _iter = jsonl_iter

    _iter.dump_to_file(in_iter_, file_name, file_mode=file_mode)



from sb_json_tools import exceptions


class JsonValidatorException(exceptions.JsonToolsException):
    """
    Base exception for json_tools.val
    """


class SchemaDefinitionException(JsonValidatorException):
    pass


class ValidationException(JsonValidatorException):
    pass



from typing import Dict
from typing import IO
from typing import Iterable
from typing import Union

import codecs
import io

from sb_json_tools import jsonlib


def dump(data: Union[Dict, Iterable], fp: IO):
    if isinstance(fp, io.BufferedIOBase):
        fp = codecs.getwriter("utf-8")(fp)

    if isinstance(data, (dict, str, int, float, bool)):
        fp.write(jsonlib.dumps(data, **jsonlib.JSON_SETTINGS))
        # fp.write("\n")
        return

    try:
        for obj in data:
            fp.write(jsonlib.dumps(obj, **jsonlib.JSON_SETTINGS))
            fp.write("\n")
    except TypeError:
        fp.write(jsonlib.dumps(data, **jsonlib.JSON_SETTINGS))
        fp.write("\n")


def load(fp: IO) -> Iterable:
    if isinstance(fp.read(0), bytes):
        fp = codecs.getreader("utf-8")(fp)
    for line in fp:
        yield jsonlib.loads(line, **jsonlib.JSON_SETTINGS)


def load_from_file(file_name: str, *, file_mode: str = None):
    if not file_mode:
        file_mode = "r"
    with open(file_name, file_mode) as fp:
        yield from load(fp)


def dump_to_file(obj, file_name, *, file_mode: str = None):
    if not file_mode:
        file_mode = "bw"
    with open(file_name, file_mode) as fp:
        dump(obj, fp)
